/*
 * Cache Class
 *
 * @public
 * @class Cache
 */
class Cache {

  /*
   * Cache constructor
   *
   * @public
   * @constructor
   * @param {Object} logger
   */
  constructor(logger) {

    this.logger = logger || console

    let redis

    if (process.env.NODE_ENV === 'test') {
      redis = require('redis-mock')
    } else {
      redis = require('redis')
    }

    this.logger.debug('Connecting to redis client', {
      port: process.env.REDIS_DB_PORT,
      host: process.env.REDIS_DB_HOST,
    })

    this.client = redis.createClient(
      process.env.REDIS_DB_PORT,
      process.env.REDIS_DB_HOST,
      {
        password: process.env.REDIS_DB_PASSWORD,
        retry_strategy: function (options) {
          // Don't ever retry to connect - we want to just immediately fail and
          // bypass the cache. We will try and open a new connection next time
          // we check the cache.
          // Returning undefined ends reconnecting with built in error.
          return undefined
        },
      },
    )

  }

  /*
   * Invalidate a cache key
   *
   * @public
   * @param {String} key - key to invalidate
   */
  invalidate(key) {

    return new Promise((resolve, reject) => {

      this.client.del(key, (err) => {

        if (err) {
          return reject(err)
        }

        resolve()

      })

    })

  }

  /*
   * Invalidate all cache keys for an application
   *
   * @public
   * @param {String} applicationId - id of application to invalidate all keys for
   */
  invalidateApplication(applicationId) {

    return new Promise((resolve, reject) => {

      this.client.keys(`${applicationId}:*`, async (err, keys) => {

        if (err) {
          return reject(err)
        }

        for (const key of keys) {

          await this.invalidate(key)

        }

        resolve()

      })

    })

  }

  /*
   * Invalidate all cache keys
   *
   * @public
   */
  invalidateAll() {

    return new Promise((resolve, reject) => {

      this.client.flushall((err) => {

        if (err) {
          return reject(err)
        }

        resolve()

      })

    })

  }

  /*
   * Get the value for a cache key, if it exists
   *
   * @public
   * @param {String} key - key to get value of
   * @returns {String|Object} - value in cache key, object if value was json
   */
  get(key) {

    return new Promise((resolve, reject) => {

      this.client.get(key, (err, value) => {

        if (err) {
          return reject(err)
        }

        try {
          const json = JSON.parse(value)
          resolve(json)
        } catch(e) {
          resolve(value)
        }

      })

    })

  }

  /*
   * Set the expiry of a cache key
   *
   * @public
   * @param {String} key - key to set expiry of
   * @param {Number} ttl - ttl in seconds
   */
  expire(key, ttl) {

    return new Promise((resolve, reject) => {

      this.client.expire(key, ttl, (err) => {

        if (err) {
          return reject(err)
        }

        resolve()

      })

    })

  }

  /*
   * Set the value of a cache key
   *
   * @public
   * @param {String} key - key to set value of
   * @param {String|Object} value - value to set cache key to
   * @param {Number} ttl - ttl in seconds
   */
  set(key, value, ttl) {

    return new Promise((resolve, reject) => {

      this.client.set(key, value, async (err) => {

        if (err) {
          return reject(err)
        }

        if (ttl) {

          await this.expire(key, ttl)

        }

        resolve()

      })

    })

  }

  /*
   * Cache the value of a request.
   *
   * We do this by rebinding the `send` method of the response object to a new
   * function that forwards the response through and also sets that response in
   * the cache.
   *
   * @public
   * @param {String} applicationId - id of application to scope the key by
   * @param {String} key - key to use for the cache
   * @param {Object} res - response object of the request
   * @param {Number} ttl - ttl in seconds
   */
  cacheRequest(applicationId, key, res, ttl) {

    const send = res.send.bind(res)

    // Wrap res.send so we can intercept and cache the response
    res.send = async (body) => {

      // Send response to client
      const response = send(body)

      try {

        // Save response to cache
        await this.set(`${applicationId}:${key}`, body, ttl)

      } catch(err) {

        this.logger.error(
          'Error setting cache', {err, applicationId, key, res, ttl}
        )

      }

      return response

    }

  }

}

module.exports = Cache
