const Cache = require('./cache')

/*
 * Cache middleware layer on rethinkdb.
 *
 * @public
 * @param {Object} ttl - how many seconds cache should stay valid for, default 10 minutes
 */
module.exports = function cache(ttl=600) {

  let cache

  return async function (req, res, next) {

    if (process.env.SKIP_CACHE) {
      return next()
    }

    const logger = req.logger || console

    // If the client has disconnected due to an error, reconnect
    if (!cache || !cache.client || !cache.client.connected) {
      cache = new Cache(logger)
    }

    const key = req.originalUrl
    const applicationId = req.headers.applicationid

    try {

      const start = new Date()
      const value = await cache.get(`${applicationId}:${key}`)
      const end = new Date()

      if (value === null) {

        // URL not found in cache, process request normally and cache response
        logger.debug('Cache Miss', {key, time: end - start})
        cache.cacheRequest(applicationId, key, res, ttl)
        next()

      } else {

        // URL found in cache, return cached response
        logger.debug('Cache Hit', {key, time: end - start})
        res.send(value)

      }

    } catch(e) {

      logger.error(`Cache Error: ${e}`)
      next()

    }

  }

}
